.style-grid {
  background: none;
}

.style-grid child {
  margin: 6px 6px 6px 6px;
}

.style-grid child:selected {
  background: none;
}

.style-grid child:hover {
  background: none;
}

.style-grid child picture {
  border-radius: 12px;
  box-shadow: 0 0 0 1px @borders;
}

.style-grid child:selected picture {
  box-shadow: 0 0 0 3px @theme_selected_bg_color;
}

.style-grid child:focus {
  outline-style: none;
}

.style-grid child:focus picture {
  outline: 0 solid transparent;
  outline-color: alpha(@theme_selected_bg_color, .5);
  outline-width: 2px;
  outline-offset: 2px;
  transition: outline-color 200ms cubic-bezier(0.25, 0.46, 0.45, 0.94),
              outline-width 200ms cubic-bezier(0.25, 0.46, 0.45, 0.94),
              outline-offset 200ms cubic-bezier(0.25, 0.46, 0.45, 0.94);
}

.style-grid child image#indicator {
  opacity: 0;
}

.style-grid child:selected image#indicator {
  color: @theme_selected_bg_color;
  opacity: 1;
}

.drag-overlay-status-page {
  background-color: color-mix(in srgb, var(--accent-bg-color), transparent);
  color: var(--accent-fg-color);
  padding: 32px;
}

.blurred {
  filter: blur(6px);
}

.drag-overlay {
  box-shadow: none;
}

.import-mode {
  background: color-mix(in srgb, currentColor 10%, transparent);
  color: inherit;
  border-radius: 99px;
  padding-top: 3px;
  padding-bottom: 3px;
  padding-left: 6px;
  padding-right: 6px;
}
